package dareka.processor;

import java.io.IOException;
import java.net.Socket;
import java.util.HashSet;
import java.util.Set;

import dareka.common.Config;

public abstract class Resource {
    /**
     * Represents resource to be sent.
     *
     */
    public enum Type {
        URL, STRING, HOSTPORT
    }

    private Set<TransferListener> listeners = new HashSet<TransferListener>();

    /**
     * Factory to create resource object. This is provided for convenience.
     * 
     * @param type
     * @param resource
     * @return
     * @throws IOException
     */
    public static Resource get(Type type, String resource) throws IOException {
        if (resource == null) {
            throw new IllegalArgumentException("resource must not be null");
        }
        
        if (type == Type.URL) {
            return new URLResource(resource);
        } else if (type == Type.STRING) {
            return new StringResource(resource);
        } else if (type == Type.HOSTPORT) {
            return new HostportResource(resource);
        } else {
            throw new IllegalArgumentException("invalid type: " + type);
        }
    }

    public void addTransferListener(TransferListener l) {
        if (l == null) {
            throw new IllegalArgumentException("TransferListener must not be null");
        }
        
        listeners.add(l);
    }

    protected int getListenersSize() {
        return listeners.size();
    }
    
    protected void fireOnResonseHeader(HttpResponseHeader responseHeader) {
        for (TransferListener l : listeners) {
            l.onResonseHeader(responseHeader);
        }
    }
    
    protected void fireOnTransferring(byte[] buf, int length) {
        for (TransferListener l : listeners) {
            l.onTransferring(buf, length);
        }
    }
    
    protected void fireOnTransferEnd(boolean completed) {
        for (TransferListener l : listeners) {
            l.onTransferEnd(completed);
        }
    }

    /**
     * Transfer this resource to specified receiver
     * 
     * @param receiver
     * @param requestHeader
     * @param config
     * @return true means it is possible to continue useing the receiver.
     * @throws IOException
     */
    public abstract boolean transferTo(Socket receiver,
            HttpRequestHeader requestHeader, Config config) throws IOException;

}
