package dareka.processor;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.Socket;
import java.net.URL;
import java.net.URLConnection;
import java.util.Map;

import dareka.common.Config;
import dareka.common.Logger;

public class URLResource extends Resource {
    private static final int BUF_SIZE = 32 * 1024;

    private URL url;

    static {
        // proxy must let browser know redirect.
        HttpURLConnection.setFollowRedirects(false);
    }

    public URLResource(String resource) throws IOException {
        url = new URL(resource);
    }

    @Override
    public boolean transferTo(Socket receiver, HttpRequestHeader requestHeader,
            Config config) throws IOException {
        return transferTo(receiver.getInputStream(),
                receiver.getOutputStream(), requestHeader, config);
    }

    public boolean transferTo(InputStream receiverIn, OutputStream receiverOut,
            HttpRequestHeader requestHeader, Config config) throws IOException {
        if (requestHeader == null) {
            requestHeader = new HttpRequestHeader("GET " + url.toString()
                    + " HTTP/1.1\r\n\r\n");
            requestHeader.setMessageHeader("User-Agent",
                    "Mozilla/4.0 (compatible; MSIE 6.0)");
        }

        Proxy proxy;

        if (config.getProxyHost().equals("")) {
            proxy = Proxy.NO_PROXY;
        } else {
            proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(config
                    .getProxyHost(), config.getProxyPort()));
        }

        URLConnection con = url.openConnection(proxy);

        prepareForConnect(requestHeader, receiverIn, con);

        con.connect();

        boolean canContinue = true;
        // ensure consuming errorStream for keep-alive.
        // see http://java.sun.com/javase/6/docs/technotes/guides/net/http-keepalive.html
        try {
            sendBodyIfNeccessary(receiverIn, requestHeader, con);

            boolean knownLengthContent = isKnownLengthContent(con);
            if (!knownLengthContent) {
                canContinue = false;
            }

            HttpResponseHeader responseHeader = getResponseHeader(con,
                    knownLengthContent);

            fireOnResonseHeader(responseHeader);

            HttpUtil.sendHeader(receiverOut, responseHeader);

            InputStream in;
            try {
                // On HttpURLConnection, even if FileNotFoundException occurred above,
                // it appears here.
                // The original FileNotFoundException is included as cause.
                in = con.getInputStream();
            } catch (IOException e) {
                // treat 401, 404 in same way as normal response. 
                // ad hoc treaing for 407
                if (con instanceof HttpURLConnection) {
                    HttpURLConnection hcon = (HttpURLConnection) con;
                    switch (hcon.getResponseCode()) {
                    case 401:
                    case 404:
                    case 407:
                        in = hcon.getErrorStream();
                        break;
                    default:
                        throw e;
                    }
                } else {
                    throw e;
                }
            }

            // below, assume IOException does not mean error response, so
            // it is not necessary to consume errorStream.
            try {
                transfer(in, receiverOut);
            } catch (IOException e) {
                Logger.debugWithThread(e);
                canContinue = false;
                try {
                    // if the data is too large, give up to cleanup.
                    if (con.getContentLength() < 100 * 1024) {
                        consumeInputStream(in);
                    }
                } catch (IOException e1) {
                    Logger.debugWithThread(e1);
                }
            } finally {
                try {
                    in.close();
                } catch (IOException e) {
                    Logger.warning(e.toString());
                }
            }
        } catch (IOException e) {
            Logger.debugWithThread(e);
            canContinue = false;
            consumeErrorStream(con);
        }

        return canContinue;
    }

    private boolean isKnownLengthContent(URLConnection con) throws IOException {
        if (con.getContentLength() == -1) {
            if (con instanceof HttpURLConnection) {
                HttpURLConnection hcon = (HttpURLConnection) con;
                int responseCode = hcon.getResponseCode();

                switch (responseCode) {
                case 204: // No Content
                case 205: // Reset Content
                case 304: // Not Modified
                    // the length is always 0. 
                    return true;
                default:
                    // do nothing
                    break;
                }
            }
            return false;
        } else {
            return true;
        }
    }

    private HttpResponseHeader getResponseHeader(URLConnection con,
            boolean knownLengthContent) throws IOException {
        HttpResponseHeader responseHeader;

        if (con instanceof HttpURLConnection) {
            responseHeader = new HttpResponseHeader(con.getHeaderField(0)
                    + "\r\n\r\n");
            copyResponseHeaderFrom(con, responseHeader);
        } else {
            responseHeader = new HttpResponseHeader("HTTP/1.1 200 OK\r\n\r\n");
            String contentType = con.getHeaderField(HttpHeader.CONTENT_TYPE);
            if (contentType != null) {
                responseHeader.setMessageHeader(HttpHeader.CONTENT_TYPE,
                        contentType);
            }
        }
        responseHeader.removeHopByHopHeaders();

        if (knownLengthContent) {
            responseHeader
                    .setMessageHeader(HttpHeader.CONNECTION, "keep-alive");

            if (con.getContentLength() != -1) {
                responseHeader.setContentLength(con.getContentLength());
            }
        } else {
            responseHeader.setMessageHeader(HttpHeader.CONNECTION, "close");
        }
        return responseHeader;
    }

    private void copyResponseHeaderFrom(URLConnection con,
            HttpResponseHeader responseHeader) throws IOException {
        int n = 1;
        for (;;) {
            String key = con.getHeaderFieldKey(n);
            if (key == null) {
                break;
            }
            responseHeader.setMessageHeader(key, con.getHeaderField(n));

            ++n;
        }
    }

    private void prepareForConnect(HttpRequestHeader requestHeader,
            InputStream receiverIn, URLConnection con) {
        if (receiverIn != null
                && (requestHeader.getContentLength() > 0 || "POST"
                        .equals(requestHeader.getMethod()))) {
            con.setDoOutput(true);
        }

        for (Map.Entry<String, String> entry : requestHeader
                .getMessageHeaders().entrySet()) {
            con.setRequestProperty(entry.getKey(), entry.getValue());
        }
    }

    private void sendBodyIfNeccessary(InputStream receiverIn,
            HttpRequestHeader header, URLConnection con) throws IOException {
        if (con.getDoOutput()) {
            OutputStream out = con.getOutputStream();
            try { // ensure out.close()
                HttpUtil.sendBody(out, receiverIn, header.getContentLength());
            } finally {
                out.close();
            }
        }
    }

    private void transfer(InputStream in, OutputStream out) throws IOException {
        byte[] buf = new byte[BUF_SIZE];
        int len;
        // split for performance
        if (getListenersSize() == 0) {
            while ((len = in.read(buf)) != -1) {
                try {
                    out.write(buf, 0, len);
                } catch (IOException e) {
                    // bad performance, but there is no chance to know
                    // write() is failed (not read()).
                    out.close();
                    throw e;
                }
            }
        } else {
            while ((len = in.read(buf)) != -1) {
                fireOnTransferring(buf, len);

                try {
                    out.write(buf, 0, len);
                } catch (IOException e) {
                    fireOnTransferEnd(false);
                    out.close();
                    throw e;
                }
            }
            fireOnTransferEnd(true);
        }
    }

    private void consumeErrorStream(URLConnection con) {
        try {
            if (con instanceof HttpURLConnection) {
                HttpURLConnection hcon = (HttpURLConnection) con;
                InputStream es = hcon.getErrorStream();
                if (es != null) {
                    try {
                        consumeInputStream(es);
                    } finally {
                        es.close();
                    }
                }
            }
        } catch (IOException e) {
            Logger.debugWithThread(e);
        }
    }

    private void consumeInputStream(InputStream in) throws IOException {
        // in case of finishing this application 
        Thread currentThread = Thread.currentThread();
 
        byte[] buf = new byte[128];
        while (!currentThread.isInterrupted() && in.read(buf) != -1) {
            // discard the input
        }
    }
}
